export module Jsonapi { 
export interface IAttributes {
    // problem with lines like
    /*
    this._receipt_number = order.attributes.receipt_number;
    this.transaction.attributes._due_date_with_format = new Date(this.transaction.attributes.due_date);
    */
    // [value: string]: boolean | string | number;
    [value: string]: any;
}


export interface ICache {
    setResource(resource: IResource): void;
    deprecateCollections(path_start_with: string): boolean;
}


export interface ICacheMemory extends ICache {
    resources: { [id: string]: IResource };

    getOrCreateCollection(url: string): ICollection;
    isCollectionExist(url: string): boolean;
    isCollectionLive(url: string, ttl: number): boolean;

    isResourceLive(id: string, ttl: number): boolean;
    getOrCreateResource(type: string, id: string): IResource;
    setCollection(url: string, collection: ICollection): void;

    removeResource(id: string): void;
}


export interface ICacheStore extends ICache {
    getResource(resource: IResource): ng.IPromise<object>;
    getCollectionFromStorePromise(url:string, includes: Array<string>, collection: ICollection): ng.IPromise<ICollection>;
    setCollection(url: string, collection: ICollection, include: Array<string>): void;
}


export interface ICollection extends Array<IResource> {
    $length: number;
    $toArray: Array<IResource>;
    $is_loading: boolean;
    $source: 'new' | 'memory' | 'store' | 'server';
    $cache_last_update: number;
    data: Array<IDataResource>; // this need disapear is for datacollection
    page: IPage;
}


export interface ICore {
    // jsonapiServices: Object;

    loadingsCounter: number;
    loadingsStart: Function;
    loadingsDone: Function;
    loadingsError: Function;
    loadingsOffline: Function;

    _register(clase: IService): boolean;
    getResourceService(type: string): IService;
    refreshLoadings(factor: number): void;
    clearCache(): void;
    duplicateResource(resource: IResource, ...relations_types: Array<string>): IResource;

    // static
    me?: IService;

    // defined on core code too
    injectedServices?: {
        $q: ng.IQService,
        JsonapiStoreService: any,
        JsonapiHttp: any,
        rsJsonapiConfig: any
    };
}


interface IDataCollection extends IDocument {
    data: Array<IDataResource>;
    page?: IPage;
    _lastupdate_time?: number;   // used when come from Store
}


interface IDataObject extends IDocument {
    data: IDataResource;
}


interface IDataResource {
    type: string;
    id: string;
    attributes?: IAttributes;
    relationships?: object;
    links?: ILinks;
    meta?: object;
}


// http://org/format/#document-top-level
interface IDocument {
    // data in child interface IJsonapiCollection
    // error in child interface IJsonapiErrors
    jsonapi?: string;
    links?: ILinks;
    included?: Array<IDataResource>;
    meta?: { [key: string]: any };
}


interface IErrors extends IDocument {
    errors: [
        {
            code?: string,
            source?: {
                attributes?: string,
                pointer: string
            },
            title?: string,
            detail?: string
        }
    ];
}


export interface IExecParamsProcessed {
    id: string;
    params: IParamsCollection | IParamsResource;
    fc_success: Function;
    fc_error: Function;
    exec_type: 'all' | 'get' | 'delete' | 'save';
}


export interface IExecParams {
    id: string;
    params?: IParamsCollection | IParamsResource | Function;
    fc_success?: Function;
    fc_error?: Function;
    exec_type: 'all' | 'get' | 'delete' | 'save';
}


// http://org/format/#document-links
export interface ILinks {
    self?: string;
    related?: {
        href: string;
        meta: object;
    };
}

export interface IPage {
    number: number;

    // http://jsonapi.org/format/#fetching-pagination
    limit?: number;

    // multinexo
    total_resources?: number;
    resources_per_page?: number;
}


interface IParamsCollection extends IParams {
    localfilter?: object;
    remotefilter?: object;
    smartfilter?: object;
    page?: IPage;
    storage_ttl?: number;
    cachehash?: string;  // solution for when we have different resources with a same id
}


interface IParamsResource extends IParams {
    id?: string;
}

export interface IParams {
    beforepath?: string;
    include?: Array<string>;
    ttl?: number;
}


interface IRelationship {
    // IDataResource added for this reason:
    // redefined from IDataResource (return errors /home/rsk/desarrollo/ts-angular-jsonapi/src/library/services/resource-functions.ts)
    data: ICollection | IResource | IDataResource | {};
}


interface IRelationships {
    [value: string]: IRelationship;
}


export interface IResource {
    type: string;
    id: string;
    attributes?: IAttributes;
    relationships: IRelationships;    // redefined from IDataResource

    is_new: boolean;
    is_loading: boolean;
    is_saving: boolean;
    lastupdate?: number;


    reset? (): void;
    addRelationship? (resource: IResource, type_alias?: string): void;
    addRelationships? (resources: ICollection, type_alias: string): void;
    removeRelationship? (type_alias: string, id: string): boolean;
    addRelationshipsArray <T extends IResource>(resources: Array<T>, type_alias?: string): void;
    save<T extends IResource>(params?: IParamsResource, fc_success?: Function, fc_error?: Function): ng.IPromise<object>;
    toObject? (params?: IParamsResource): IDataObject;
    getService(): IService;
}


export interface IResourcesById {
    [resource_id: string]: IResource;
}


export interface IResourcesByType {
    [type: string]: IResourcesById;
}

export interface ISchema {
    attributes?: object;
    relationships?: object;
    ttl?: number;
    path?: string;
}


export interface IService {
    type: string;
    schema: ISchema;
    getPrePath(): string;
    getPath(): string;
    register(): boolean;
    get<T extends IResource>(id: string | number, params?: IParamsResource | Function, fc_success?: Function, fc_error?: Function): T;
    all(params?: IParamsCollection | Function, success?: Function, error?: Function): ICollection;
    delete (id: String, params?: IParamsResource | Function, success?: Function, error?: Function): void;
    getService<T extends IService> ():T;
    clearCacheMemory? (): boolean;
    new?<T extends IResource>(): T;
    cachememory: ICacheMemory;
    cachestore: ICacheStore;
    parseFromServer(attributes: IAttributes): void;
    parseToServer?(attributes: IAttributes): void;
}

export interface IStoreObject {
    _lastupdate_time: number;
}
}